#include <iostream>
#include <fstream>
#include <sstream>
#include <cmath>
#include <string>
#include <array>
#include <unordered_map>
#include <yaml-cpp/yaml.h>

#include <TFile.h>
#include <TTree.h>
#include <TMath.h>
#include <TSystem.h>
#include <TArtEventStore.hh>
#include <TArtRawEventObject.hh>
#include <TArtCore.hh>

#include "TCordferBranchManager.hh"
#include "CalibrationManager.hh"

using namespace std;


void TCordferBranchManager::Readyaml(TString yamlfile){
  FileStat_t info;
  if(gSystem->GetPathInfo(yamlfile.Data(), info) != 0){
    cerr << "EROOR: File '" << yamlfile.Data() << "' does not exist." << endl;
    exit(0);
  }
  recipe = YAML::LoadFile(yamlfile.Data());
}


void TCordferBranchManager::CreateRawTree(){
  TString rawname = recipe["raw_tree"]["name"].as<string>();
  TString rawdesc = recipe["raw_tree"]["description"].as<string>();

  tree = new TTree(rawname, rawdesc);
}


void TCordferBranchManager::CreateMapTree(){
  TString mapname = recipe["map_tree"]["name"].as<string>();
  TString mapdesc = recipe["map_tree"]["description"].as<string>();

  mtree = new TTree(mapname, mapdesc);
}


void TCordferBranchManager::CreateCalibTree(){
  TString calibname = recipe["calib_tree"]["name"].as<string>();
  TString calibdesc = recipe["calib_tree"]["description"].as<string>();

  ctree = new TTree(calibname, calibdesc);
}


void TCordferBranchManager::CreateRawBranch(){
  YAML::Node setup = recipe["setup"].as<YAML::Node>();
  for(Int_t i=0; i<setup.size(); i++){
    Int_t seg_id = setup[i]["segment_no."].as<int>();
    TString bname = setup[i]["name"].as<string>();
    if(bname != "mhtdc"){
      Int_t nch = setup[i]["channel"].as<int>();
      TCordferRawBranch *raw_branch = new TCordferRawBranch(bname, nch);
      raw_branch->Setgeo(seg_id);
      raw_branch->MemorySet();

      TString brancharray;
      brancharray.Form("%s[%d]/I", bname.Data(), raw_branch->size);

      tree->Branch(bname.Data(), raw_branch->bra_val, brancharray);
      branch.push_back(raw_branch);
    }else{
      TString mhtdc_name = setup[i]["branches"][0]["name"].as<string>();
      Int_t nch = setup[i]["branches"][0]["channel"].as<int>();
      Int_t nhit = setup[i]["branches"][0]["maxhit"].as<int>();

      TCordferRawBranch *mhtdc_branch = new TCordferRawBranch(mhtdc_name, nch, nhit);
      mhtdc_branch->Setgeo(seg_id);
      mhtdc_branch->TwoDimMemorySet();

      TString brancharray;
      brancharray.Form("%s[%d][%d]/I", mhtdc_name.Data(), mhtdc_branch->size, mhtdc_branch->nhit);

      tree->Branch(mhtdc_name.Data(), mhtdc_branch->mhtdc_val, brancharray);
      branch.push_back(mhtdc_branch);

      TString nhit_name = setup[i]["branches"][1]["name"].as<string>();
      Int_t ch = setup[i]["branches"][1]["channel"].as<int>();
      TCordferRawBranch *nhit_branch = new TCordferRawBranch(nhit_name, ch);
      nhit_branch->Setgeo(-1);
      nhit_branch->MemorySet();

      TString nhitbrancharray;
      nhitbrancharray.Form("%s[%d]/I", nhit_name.Data(), nhit_branch->size);

      tree->Branch(nhit_name.Data(), nhit_branch->bra_val, nhitbrancharray);
      branch.push_back(nhit_branch);
    }
  }
}

void TCordferBranchManager::CreateMapBranch(){
  YAML::Node map_setup = recipe["map_setup"].as<YAML::Node>();
  for(Int_t i=0; i<map_setup.size(); i++){
    TString mname = map_setup[i]["mapfile"].as<string>();

    FileStat_t info;
    if(gSystem->GetPathInfo(mname.Data(), info) != 0){
      cerr << "EROOR: in yaml File, '" << mname.Data() << "' does not exist." << endl;
      exit(0);
    }

    ChannelMap *mfile = new ChannelMap(mname);
    for(Int_t j=0; j<mfile->detector_name.size(); j++){
      TString bname = mfile->detector_name[j];
      TCordferMapBranch *map_branch = new TCordferMapBranch(bname, mfile->segment_id[j].size());
      map_branch->CopyMapInfo(mfile->segment_id[j], mfile->channel[j]);
      map_branch->MemorySet();

      TString brancharray;
      brancharray.Form("%s[%d]/I", bname.Data(), map_branch->size);

      mtree->Branch(bname.Data(), map_branch->bra_val, brancharray);
      mbranch.push_back(map_branch);
    }
  }
}

void TCordferBranchManager::CreateCalibBranch(){
  YAML::Node map_setup = recipe["map_setup"].as<YAML::Node>();
  for(Int_t i=0; i<map_setup.size(); i++){
    TString det_name = map_setup[i]["name"].as<string>();

    //**************
    if(det_name == "ppac"){ continue; }
    //**************

    TString cname[2];
    cname[0] = map_setup[i]["calibfile"]["energy"].as<string>();
    cname[1] = map_setup[i]["calibfile"]["timing"].as<string>();

    for(Int_t j=0; j<2; j++){
      FileStat_t info;
      if(cname[j] == ""){ continue; }
      else if(gSystem->GetPathInfo(cname[j].Data(), info) != 0){
        cerr << "WARNING: in yaml File, '" << cname[j].Data() << "' does not exist." << endl;
        cerr << "         not include " << cname[j].Data() << " in ctree" << endl;
        continue;
      }

      CalibParamReader *cfile = new CalibParamReader(det_name, cname[j]);
      for(Int_t k=0; k<cfile->branch_name.size(); k++){
        TString bname = cfile->branch_name[k];
        TCordferCalibBranch *calib_branch = new TCordferCalibBranch(bname);
        calib_branch->CopyCalibInfo(cfile->dim, cfile->parameters[k]);

        TString brancharray;
        if(j==0){
          brancharray.Form("%s/D", bname.Data());
          ctree->Branch(bname.Data(), &calib_branch->cbra_val, brancharray);
          calib_branch->adc = true;
        }else{
          brancharray.Form("%s[12]/D", bname.Data());
          ctree->Branch(bname.Data(), calib_branch->cbra_tval, brancharray);
          calib_branch->adc = false;
        }
        cbranch.push_back(calib_branch);
      }
    }
    if(det_name == "mssd"){
      TString pos_name = map_setup[i]["calibfile"]["position"].as<string>();
      FileStat_t info;
      if(gSystem->GetPathInfo(pos_name.Data(), info) != 0){
        cerr << "ERROR: in yaml File, '" << pos_name.Data() << "' does not exist." << endl;
        cerr << "       mssd calibration requires " << pos_name.Data() << " !" << endl;
        exit(0);
      }

      CalibParamReader *cfile = new CalibParamReader();
      cfile->ReadMUXposCalib(pos_name);
      mpos = cfile->position_prms;
    }
  }
}


void TCordferBranchManager::CreateRawMapping(){
  for(Int_t i=0; i<branch.size(); i++){
    Int_t seg = branch[i]->geo;
    mapping[seg] = i;
  }
}


void TCordferBranchManager::CreateMapping(){
  for(Int_t i=0; i<mbranch.size(); i++){
    for(Int_t j=0; j<mbranch[i]->size; j++){
      Int_t seg = mbranch[i]->segment_id[j]*1000 + mbranch[i]->channel[j];
      Int_t index = i*1000 + j;
      mmapping[seg] = index;
    }
  }
}


void TCordferBranchManager::CreateCalibMapping(){
  Int_t mssd_tune = 0;
  for(Int_t i=0; i<cbranch.size(); i++){
    string cbranch_name = string(cbranch[i]->GetBranchName().Data());

    //for rf data
    if(cbranch_name.find("RF1") != string::npos){
      for(Int_t j=0; j<mbranch.size(); j++){
        string mbranch_name = string(mbranch[j]->GetBranchName().Data());
        if(mbranch_name.find("RF") != string::npos || mbranch_name.find("rf") != string::npos){
          cbranch[i]->geo = mbranch[j]->segment_id[0];
          cbranch[i]->ch = mbranch[j]->channel[0];
        }
      }
    }else if(cbranch_name.find("RF2") != string::npos){
      for(Int_t j=0; j<mbranch.size(); j++){
        string mbranch_name = string(mbranch[j]->GetBranchName().Data());
        if(mbranch_name.find("RF") != string::npos || mbranch_name.find("rf") != string::npos){
          cbranch[i]->geo = mbranch[j]->segment_id[1];
          cbranch[i]->ch = mbranch[j]->channel[1];
        }
      }
    }

    //for coin data
    else if(cbranch_name.find("coin") != string::npos){
      for(Int_t j=0; j<mbranch.size(); j++){
        string mbranch_name = string(mbranch[j]->GetBranchName().Data());
        if(cbranch_name == mbranch_name){
          cbranch[i]->geo = mbranch[j]->segment_id[0];
          cbranch[i]->ch = mbranch[j]->channel[0];
        }
      }
    }

    //for mssd
    else if(cbranch_name.find("mssd") != string::npos && cbranch_name.substr(0, 2) != "t_"){
      mssd_tune++;
      if(mssd_tune%16 == 1){
        Int_t mux_id = mssd_tune/16;
        cbranch[i]->mpos1_prm = mpos[2*mux_id];
        cbranch[i]->mpos2_prm = mpos[2*mux_id + 1];

        Int_t seg0, seg1, seg2, seg3;

        for(Int_t j=0; j<mbranch.size(); j++){
          string mbranch_name = string(mbranch[j]->GetBranchName().Data());
          if(cbranch_name.find(mbranch_name) != string::npos){
            seg0 = mbranch[j]->segment_id[0]*1000 + mbranch[j]->channel[0];
            cmapping[seg0] = i;
            muxmapping[seg0] = 1;

            seg1 = mbranch[j]->segment_id[1]*1000 + mbranch[j]->channel[1];
            cmapping[seg1] = i;
            muxmapping[seg1] = 2;

            seg2 = mbranch[j]->segment_id[2]*1000 + mbranch[j]->channel[2];
            cmapping[seg2] = i;
            muxmapping[seg2] = 3;

            seg3 = mbranch[j]->segment_id[3]*1000 + mbranch[j]->channel[3];
            cmapping[seg2] = i;
            muxmapping[seg2] = 4;
          }
        }
        //cout << i << " " << cbranch_name << " " << seg0 << ", " << seg1 << ", " << seg2 << ", " << seg3 << endl;
        continue;
      }else{
        //cout << i << " " << cbranch_name << endl;
        continue;
      }
    }else if(cbranch_name.find("mssd") != string::npos && cbranch_name.substr(0, 2) == "t_"){
      for(Int_t j=0; j<mbranch.size(); j++){
        string mbranch_name = string(mbranch[j]->GetBranchName().Data());
        if(cbranch_name.find(mbranch_name) != string::npos){
          cbranch[i]->geo = mbranch[j]->segment_id[4];
          cbranch[i]->ch = mbranch[j]->channel[4];
        }
      }
    }

    //for ppac
    //else if(){}

    //for dssd, ssd data
    else{
      bool tdc_flag = false;
      if(cbranch_name.substr(0, 2) == "t_"){
        cbranch_name.erase(0, 2);
        tdc_flag = true;
      }
      for(Int_t j=0; j<mbranch.size(); j++){
        string mbranch_name = string(mbranch[j]->GetBranchName().Data());
        if(cbranch_name == mbranch_name){
          if(tdc_flag){
            cbranch[i]->geo = mbranch[j]->segment_id[1];
            cbranch[i]->ch = mbranch[j]->channel[1];
          }else{
            cbranch[i]->geo = mbranch[j]->segment_id[0];
            cbranch[i]->ch = mbranch[j]->channel[0];
          }
        }
      }
    }

    Int_t seg = cbranch[i]->geo*1000 + cbranch[i]->ch;
    Int_t index = i;
    cmapping[seg] = index;

    //cout << i << " " << cbranch_name << " " << seg << endl;
  }
}


void TCordferBranchManager::GiveRawValue(Int_t geo, Int_t ch, Int_t val){
  //if not register segment
  auto itr = mapping.find(geo);
  if(itr == mapping.end()){
    if(rmapflag == false){
      cout << "WARNING: rmapping is not completed" << endl;
    }
    rmapflag=true;
    return;
  }

  Int_t index = mapping[geo];

  if(branch[index]->GetBranchName() == "mhtdc"){
    Int_t ht_id = branch[mapping[-1]]->bra_val[ch];
    branch[index]->mhtdc_val[ch][ht_id] = val;
    branch[mapping[-1]]->bra_val[ch] += 1;
  }else{
    branch[index]->bra_val[ch] = val;
  }
}


void TCordferBranchManager::GiveMapValue(Int_t geo, Int_t ch, Int_t val){
  Int_t seg = geo*1000 + ch;

  //if not register segment
  auto itr = mmapping.find(seg);
  if(itr == mmapping.end()){
    if(mapflag == false){
      cout << "WARNING: mapping is not completed" << endl;
    }
    mapflag=true;
    return;
  }

  Int_t index = mmapping[seg];
  Int_t j = index % 1000;
  Int_t i = (index - j)/1000;

  mbranch[i]->bra_val[j] = val;
}


void TCordferBranchManager::GiveCalibValue(Int_t geo, Int_t ch, Int_t val){
  Int_t seg = geo*1000 + ch;

  //if not register segment
  auto itr = cmapping.find(seg);
  if(itr == cmapping.end()){
    if(cmapflag == false){
      cout << "WARNING: cmapping is not completed" << endl;
    }
    cmapflag=true;
    return;
  }

  Int_t index = cmapping[seg];
  string cbranch_name = string(cbranch[index]->GetBranchName().Data());

  //for mssd energy calibration
  if(cbranch_name.find("mssd") != string::npos && cbranch_name.substr(0, 2) != "t_"){
    Int_t mux_sig = muxmapping[seg];
    if(mux_sig==1 && cbranch[index]->mux_pos1==false){
      cbranch[index]->SetGeometrytmp1(geo, ch, val);
      cbranch[index]->mux_status1=true;

    }else if(mux_sig==1 && cbranch[index]->mux_pos1==true){
      if(cbranch[index]->mux1_junk == true){ return; }
      Int_t add_ch = cbranch[index]->mux_ch1;
      cbranch[index + add_ch]->cbra_val = GetCalibValue(val, cbranch[index+add_ch]->calib_param);

    }else if(mux_sig==2 && cbranch[index]->mux_pos2==false){
      cbranch[index]->SetGeometrytmp2(geo, ch, val);
      cbranch[index]->mux_status2=true;

    }else if(mux_sig==2 && cbranch[index]->mux_pos2==true){
      if(cbranch[index]->mux2_junk == true){ return; }
      Int_t add_ch = cbranch[index]->mux_ch2;
      cbranch[index + add_ch]->cbra_val = GetCalibValue(val, cbranch[index+add_ch]->calib_param);

    }else if(mux_sig==3){
      cbranch[index]->mux_pos1 = true;
      for(Int_t i=0; i<16; i++){
        if(cbranch[index]->mpos1_prm[i] < val && cbranch[index]->mpos1_prm[i+1] > val){
          cbranch[index]->mux_ch1 = i;
        }
      }
      if(cbranch[index]->mux_ch1 < 0){ cbranch[index]->mux1_junk = true; }

    }else if(mux_sig==4){
      cbranch[index]->mux_pos2 = true;
      for(Int_t i=0; i<16; i++){
        if(cbranch[index]->mpos2_prm[i] < val && cbranch[index]->mpos2_prm[i+1] > val){
          cbranch[index]->mux_ch2 = i;
        }
      }
      if(cbranch[index]->mux_ch2 < 0){ cbranch[index]->mux2_junk = true; }
    }

  //for others
  }else{
    if(cbranch[index]->adc){
      cbranch[index]->cbra_val = GetCalibValue(val, cbranch[index]->calib_param);
    }else{
      cbranch[index]->cbra_tval[cbranch[index]->count_hit] = GetCalibValue(val, cbranch[index]->calib_param);
      cbranch[index]->count_hit++;
    }
  }
}


void TCordferBranchManager::GiveWaitCalibValue(){
  for(Int_t i=0; i<cbranch.size(); i++){
    if(cbranch[i]->mux_status1==true){ GiveCalibValue(cbranch[i]->geo_tmp1, cbranch[i]->ch_tmp1, cbranch[i]->val_tmp1); }
    if(cbranch[i]->mux_status2==true){ GiveCalibValue(cbranch[i]->geo_tmp2, cbranch[i]->ch_tmp2, cbranch[i]->val_tmp2); }
  }
}


void TCordferBranchManager::ClearRawAll(){
  for(Int_t i=0; i<branch.size(); i++){
    if(branch[i]->GetBranchName() == "mhtdc"){
      for(Int_t j=0; j<branch[i]->size; j++){
        for(Int_t k=0; k<branch[i]->nhit; k++){
          branch[i]->mhtdc_val[j][k] = -500;
        }
      }
    }else if(branch[i]->GetBranchName() == "mhtdc_nhit"){
      for(Int_t j=0; j<branch[i]->size; j++){
        branch[i]->bra_val[j] = 0;
      }
    }else{
      for(Int_t j=0; j<branch[i]->size; j++){
        branch[i]->bra_val[j] = -500;
      }
    }
  }
}


void TCordferBranchManager::ClearMapAll(){
  for(Int_t i=0; i<mbranch.size(); i++){
    for(Int_t j=0; j<mbranch[i]->size; j++){
      mbranch[i]->bra_val[j] = -500;
    }
  }
}


void TCordferBranchManager::ClearCalibAll(){
  for(Int_t i=0; i<cbranch.size(); i++){
    cbranch[i]->mux_status1=false;
    cbranch[i]->mux_status2=false;
    cbranch[i]->mux_pos1=false;
    cbranch[i]->mux_pos2=false;
    cbranch[i]->mux1_junk=false;
    cbranch[i]->mux2_junk=false;
    cbranch[i]->mux_ch1=-99;
    cbranch[i]->mux_ch2=-99;
    cbranch[i]->geo_tmp1=-99;
    cbranch[i]->geo_tmp2=-99;
    cbranch[i]->ch_tmp1=-99;
    cbranch[i]->ch_tmp2=-99;
    cbranch[i]->val_tmp1=-99;
    cbranch[i]->val_tmp2=-99;

    if(cbranch[i]->adc){
      cbranch[i]->cbra_val = -500.5;
    }else{
      for(Int_t j=0; j<12; j++){
        cbranch[i]->cbra_tval[j] = -500.5;
      }
      cbranch[i]->count_hit = 0;
    }
  }
}

void TCordferBranchManager::FileClose(){
  rootfile->Close();
  for(Int_t i=0; i<branch.size(); i++){
    if(branch[i]->GetBranchName() == "mhtdc"){
      branch[i]->FreeTwoDimMemory();
    }else{
      branch[i]->FreeMemory();
    }
  }
  for(Int_t i=0; i<mbranch.size(); i++){
    mbranch[i]->FreeMemory();
  }
}

// for debug
void TCordferBranchManager::DEBUG_PRINT(){
  for(Int_t i=0; i< branch.size(); i++){
    cout << "\tDEBUG: " << branch[i]->GetBranchName() << endl;
    cout << "\t\t" << branch[i]->size << " " << branch[i]->nhit << endl;
  }
}
